#ifndef SCREENCAST_H
#define SCREENCAST_H

#include "screencast_common.h"

void xdpw_screencast_instance_destroy(struct xdpw_screencast_instance *cast);
void xdpw_screencast_instance_teardown(struct xdpw_screencast_instance *cast);

#endif
